﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Reflection;
using log4net;
using Lync.Widget.Service.Common;
using Microsoft.Rtc.Collaboration;
using Microsoft.Rtc.Signaling;

namespace Lync.Widget.Service.InstantMessaging
{
	public class UCInstantMessageRouter
	{
		private static readonly ILog Log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

		private static UCInstantMessageRouter _instance;
		private static CollaborationPlatform _collabPlatform;
		private List<UCInstantMessageProvider> _providers = new List<UCInstantMessageProvider>();
		private List<MessageQueue> _messageQueues = new List<MessageQueue>();

		private UCInstantMessageRouter()
		{
            var transportType = SipTransportType.Tls;

			var clientPlatformSettings = new ClientPlatformSettings(ConfigurationManager.AppSettings["ApplicationName"], transportType);
            _collabPlatform = new CollaborationPlatform(clientPlatformSettings);

            _collabPlatform.BeginStartup(EndPlatformStartup, _collabPlatform);
		}

		public static UCInstantMessageRouter Instance
		{
			get
			{
				if (_instance == null)
				{
					_instance = new UCInstantMessageRouter();
				}
				return _instance;
			}
		}

        private void EndPlatformStartup(IAsyncResult ar)
        {
            var collabPlatform = ar.AsyncState as CollaborationPlatform;
            try
            {
                // The platform should now be started.
                collabPlatform.EndStartup(ar);
                // It should be noted that all the re-thrown exceptions will crash the application. This is intentional.
                // Ideal exception handling would report the error and shut down nicely. In production code, consider using 
                // an IAsyncResult implementation to report the error instead of throwing or put the implementation
                // in this try block.
            }
            catch (OperationFailureException opFailEx)
            {
                // OperationFailureException will be thrown when the platform cannot establish, here, usually due to invalid data.
                Log.Error("OperationFailureException", opFailEx);
                throw;
            }
            catch (ConnectionFailureException connFailEx)
            {
                // ConnectionFailureException will be thrown when the platform cannot connect.
                // ClientPlatforms will not throw this exception on startup.
				Log.Error("ConnectionFailureException", connFailEx);
                throw;
            }
            catch (RealTimeException realTimeEx)
            {
                // RealTimeException may be thrown as a result of any UCMA operation.
				Log.Error("RealTimeException", realTimeEx);
                throw;
            }
        }

		public void InitChatSession(string session, string employeeSip, string username, string page, string topic)
		{

			//check if session does not exist yet.
			if (QueueExists(session))
			{
				//throw an error because this could lead to a conversation hack
				Log.Error(string.Format("cache for {0} was being initiated but it already exists.", session));
				throw new InvalidOperationException(string.Format("cache for {0} was being initiated but it already exists.", session));
			}

			var queue = new MessageQueue(session);
			_messageQueues.Add(queue);

			//check if session does not exist yet.
			if (SessionProviderExists(session))
			{
				//throw an error because this could lead to a conversation hack
				Log.Error(string.Format("session {0} was being initiated but it already exists.", session));
				throw new InvalidOperationException(string.Format("session {0} was being initiated but it already exists.", session));
			}

			//create provider for the session
			var provider = new UCInstantMessageProvider(_collabPlatform, session, employeeSip, username, page, topic);
			provider.ChatRequestAcceptanceChanged += provider_ChatRequestAcceptanceChanged;
			provider.ChatMessageReceivedFromLyncCLient += provider_ChatMessageReceivedFromLyncClient;
			provider.InstantMessageProviderExceptionOccured += provider_InstantMessageProviderExceptionOccured;
            provider.Notification += provider_Notification;
			provider.ChatClosed += provider_ChatClosed;
			provider.ComposingStateChanged += provider_ComposingStateChanged;
			
			//add provider to provider collection
			_providers.Add(provider);
		}

		void provider_ComposingStateChanged(object sender, ComposingStateChangedEventArgs e)
		{
			var provider = (UCInstantMessageProvider)sender;
			QueueMessage(provider.Session, new Message(e.composingState.ToString(), MessageType.ComposingStateChanged));
		}

		void DisposeProvider(UCInstantMessageProvider provider)
		{
			provider.ChatRequestAcceptanceChanged -= provider_ChatRequestAcceptanceChanged;
			provider.ChatMessageReceivedFromLyncCLient -= provider_ChatMessageReceivedFromLyncClient;
			provider.InstantMessageProviderExceptionOccured -= provider_InstantMessageProviderExceptionOccured;
			provider.Notification -= provider_Notification;
			provider.ChatClosed -= provider_ChatClosed;
			provider.ComposingStateChanged -= provider_ComposingStateChanged;
			
			provider.Dispose();
			_providers.Remove(provider);
		}

		void provider_ChatClosed(object sender, EventArgs e)
		{
			var provider = (UCInstantMessageProvider)sender;
			QueueMessage(provider.Session, new Message("Sorry, this user is no longer available for chat. You can send an email or place a call instead.", MessageType.ChatClosed));
			DisposeProvider(provider);
		}

        /// <summary>
        /// Occurs when the provider sends a notification to pass to the end user
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void provider_Notification(object sender, NotificationEventArgs e)
        {
			var provider = (UCInstantMessageProvider)sender;
			QueueMessage(provider.Session, new Message(e.message, MessageType.ServiceNotification));
        }

		/// <summary>
		/// Occurs when the provider notifies the router of an internal exception.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void provider_InstantMessageProviderExceptionOccured(object sender, InstantMessageProviderExceptionOccuredEventArgs e)
		{
			var provider = (UCInstantMessageProvider) sender;
			QueueMessage(provider.Session, new Message(e.exception.Message, MessageType.ServiceError));
		}

		/// <summary>
		/// Occurs when the provider notifies the router that a chat message was received from the lync client
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void provider_ChatMessageReceivedFromLyncClient(object sender, MessageReceivedFromLyncClientEventArgs e)
		{
			var provider = (UCInstantMessageProvider)sender;
			QueueMessage(provider.Session, new Message(e.message, MessageType.InstantMessage));
		}

		/// <summary>
		/// Occurs when the provider notifies the router that a lync client has accepted a chat request
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		void provider_ChatRequestAcceptanceChanged(object sender, ChatRequestAcceptanceChangedEventArgs e)
		{
			var provider = (UCInstantMessageProvider) sender;
			if (e.isAccepted)
			{
				QueueMessage(provider.Session, new Message("Chat request accepted by the user! Go ahead and say something.", MessageType.ChatRequestAccepted));
			}
			else
			{
				QueueMessage(provider.Session, new Message("Sorry, this user is no longer available for chat. You can send an email or place a call instead.", MessageType.ChatRequestDenied));
				DisposeProvider(provider);
			}
		}

		void QueueMessage(string session, Message message)
		{
			MessageQueue cache;
			try
			{
				cache = Getcache(session);
			}
			catch (QueueNotFoundException cacheNotFoundException)
			{
				Log.WarnFormat("QueueNotFoundException for session {0} and message '{1}'", session, message);
				throw;
			}

			cache.Add(message);
		}

		public void SendInstantMessageToLyncClient(string session, string message)
		{
			UCInstantMessageProvider provider;
			try
			{
				provider = GetProvider(session);
			}
			catch (SessionNotFoundException sessionNotFoundException)
			{
				Log.WarnFormat("SessionNotFoundException for session {0} and message '{1}'", session, message);
				throw;
			}

			provider.SendInstantMessage(message);
		}

		public List<Message> GetInstantMessagesFromLynClient(string session)
		{
			MessageQueue cache;
			try
			{
				cache = Getcache(session);
			}
			catch (QueueNotFoundException cacheNotFoundException)
			{
				Log.WarnFormat("QueueNotFoundException for session {0}", session);
				throw;
			}

			return cache.GetAllAndDelete();
		}

		private UCInstantMessageProvider GetProvider(string session)
		{
			var provider = _providers.SingleOrDefault(p => p.Session.Equals(session));
			if (provider != null)
			{
				return provider;
			}
			throw new SessionNotFoundException();
		}

		private MessageQueue Getcache(string session)
		{
			var cache = _messageQueues.SingleOrDefault(p => p.Session.Equals(session));
			if (cache != null)
			{
				return cache;
			}
			throw new QueueNotFoundException();
		}

		private bool SessionProviderExists(string session)
		{
			var provider = _providers.SingleOrDefault(p => p.Session.Equals(session));
			return provider != null;
		}

		private bool QueueExists(string session)
		{
			var cache = _messageQueues.SingleOrDefault(p => p.Session.Equals(session));
			return cache != null;
		}

		public ServiceHelpers.StatusReport GetStatusReport()
		{
			return new ServiceHelpers.StatusReport
			                   	{
			                   		NumberOfMessageQueues = _messageQueues.Count,
									TotalNumberOfMessages = _messageQueues.Sum(queue => queue.GetMessageCount()),
									NumberOfProviders = _providers.Count,
			                   	};
		}
	}
}