﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using log4net;
using Lync.Widget.Service.Common;
using Microsoft.Rtc.Collaboration.Presence;

namespace Lync.Widget.Service.Presence
{
	/// <summary>
	/// The default UCMA implementation of IPresenceCache.
	/// Gathers and caches presence updates for the requested sip uri(s)
	/// </summary>
	public class UcPresenceCache: IPresenceCache
	{
		private static readonly ILog Log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

		private Dictionary<string, PresenceSubscription> m_presenceSubscriptionsBySipUri =
			new Dictionary<string, PresenceSubscription>(StringComparer.CurrentCultureIgnoreCase);
		private object m_syncObject = new object();
		private Timer _staleTimer;
		private DateTime StaleTime { get { return DateTime.Now - StaleInterval; } }

		/// <summary>
		/// When set to true, unsubscribe and resubscribe after every time we received presence information.
		/// Implemented because the presence information would only be given once after subscription in our environment.
		/// </summary>
		private bool circumventSubscription = true;

		private TimeSpan resubscribeDelay = TimeSpan.FromSeconds(4);

		private IPresenceProvider m_presenceProvider;

		private Dictionary<string, Timer> m_presenceTimersBySipUri = new Dictionary<string, Timer>(StringComparer.CurrentCultureIgnoreCase);

		/// <summary>
		/// The interval to run cache cleanup. If any presence subscriptions have not been requested in the last stale interval
		/// they will be removed from the cache
		/// </summary>
		public TimeSpan StaleInterval { get; set; }

		public UcPresenceCache(IPresenceProvider provider)
		{
			m_presenceProvider = provider;
			m_presenceProvider.PresenceChanged += PresenceProvider_PresenceChanged;
			StaleInterval = TimeSpan.FromHours(2);
			_staleTimer = new Timer(OnStaleTimerTick, null, StaleInterval, TimeSpan.FromMilliseconds(-1));
		}

		public PresenceSubscription GetPresence(string sipUri)
		{
			lock (m_syncObject)
			{
				if (!m_presenceSubscriptionsBySipUri.ContainsKey(sipUri))
				{
					m_presenceSubscriptionsBySipUri.Add(sipUri, new PresenceSubscription
					{
						Availability = (long)PresenceAvailability.None,
						ContactName = "Unknown",
						SipUri = sipUri,
						Status = "Unknown",
						SubscriptionCreated = DateTime.Now
					});
					m_presenceProvider.SubscribeToPresence(sipUri);
				}
				//update the timestamp whenever the presence is requested
				//this prevents a presence from being cleared out of the cache at the StaleInterval
				m_presenceSubscriptionsBySipUri[sipUri].SubscriptionCreated = DateTime.Now;
				return m_presenceSubscriptionsBySipUri[sipUri];
			}
		}

		public List<PresenceSubscription> GetPresence(List<string> sipUris)
		{
			List<PresenceSubscription> presences = new List<PresenceSubscription>();
			foreach (string sipUri in sipUris)
			{
				presences.Add(GetPresence(sipUri));
			}
			return presences;
		}

		#region Manage Presence Changes
		private void PresenceProvider_PresenceChanged(object sender, PresenceChangedEventArgs e)
		{
			lock (m_syncObject)
			{
				foreach (PresenceSubscription presence in e.Presences)
				{
					if (m_presenceSubscriptionsBySipUri.ContainsKey(presence.SipUri))
					{
						UpdatePresenceSubscriptionWithChanges(presence);

						if (circumventSubscription)
						{
							m_presenceProvider.UnsubscribeToPresence(presence.SipUri);
							if (m_presenceTimersBySipUri.ContainsKey(presence.SipUri))
							{
								m_presenceTimersBySipUri.Remove(presence.SipUri);
							}

							Timer timer = new Timer(o => m_presenceProvider.SubscribeToPresence(presence.SipUri), null, resubscribeDelay, TimeSpan.FromMilliseconds(-1));
							m_presenceTimersBySipUri.Add(presence.SipUri, timer);
						}

					}
					else //received a presence changed event for an expired subscription
					{
						Log.Info(string.Format("Ignoring presence update for sipUri {0} since there are no active subscriptions for this uri", presence.SipUri));
					}
				}
			}
		}

		private void UpdatePresenceSubscriptionWithChanges(PresenceSubscription presence)
		{
			if (presence.Availability != PresenceSubscription.NoChangeInAvailability)
			{
				m_presenceSubscriptionsBySipUri[presence.SipUri].Availability = presence.Availability;
			}
			if (presence.ContactName != PresenceSubscription.NoChange)
			{
				m_presenceSubscriptionsBySipUri[presence.SipUri].ContactName = presence.ContactName;
			}
			if (presence.Status != PresenceSubscription.NoChange)
			{
				m_presenceSubscriptionsBySipUri[presence.SipUri].Status = presence.Status;
			}
			if (presence.Email != PresenceSubscription.NoChange)
			{
				m_presenceSubscriptionsBySipUri[presence.SipUri].Email = presence.Email;
			}
			if (presence.Phone != PresenceSubscription.NoChange)
			{
				m_presenceSubscriptionsBySipUri[presence.SipUri].Phone = presence.Phone;
			}
			if (presence.Mobile != PresenceSubscription.NoChange)
			{
				m_presenceSubscriptionsBySipUri[presence.SipUri].Mobile = presence.Mobile;
			}
			if (presence.JobTitle != PresenceSubscription.NoChange)
			{
				m_presenceSubscriptionsBySipUri[presence.SipUri].JobTitle = presence.JobTitle;
			}

			m_presenceSubscriptionsBySipUri[presence.SipUri].CanChat = CanPresenceChat(presence);
		}

		private bool CanPresenceChat(PresenceSubscription presence)
		{
			return (presence.Availability >= 3000 && presence.Availability <= 4499)
				|| (presence.Availability >= 6000 && presence.Availability <= 7499);
		}
		#endregion

		#region Cache Cleanup
		private void RemoveStaleSubscriptions()
		{
			lock (m_syncObject)
			{
				DateTime staleTime = StaleTime;
				var staleKvps = m_presenceSubscriptionsBySipUri.Where(kvp => kvp.Value.SubscriptionCreated <= StaleTime).ToList();
				foreach (var staleKvp in staleKvps)
				{
					string sipUri = staleKvp.Key;
					Log.Info(string.Format("Removing stale subscription for sip uri {0}, last requested at {1}", sipUri, staleKvp.Value.SubscriptionCreated));
					m_presenceProvider.UnsubscribeToPresence(sipUri);
					m_presenceSubscriptionsBySipUri.Remove(sipUri);
					m_presenceTimersBySipUri.Remove(sipUri);
				}
			}
		}

		private void OnStaleTimerTick(object state)
		{
			//stop the stale timer
			_staleTimer.Change(TimeSpan.FromMilliseconds(-1), TimeSpan.FromMilliseconds(-1));
			RemoveStaleSubscriptions();
			_staleTimer.Change(StaleInterval, TimeSpan.FromMilliseconds(-1));
		}
		#endregion

		public void Reset(string sipUri)
		{
			m_presenceProvider.UnsubscribeToPresence(sipUri);
			m_presenceSubscriptionsBySipUri.Remove(sipUri);
		}

		public void ResetAll()
		{
			foreach (var sipUri in m_presenceSubscriptionsBySipUri.Keys)
			{
				m_presenceProvider.UnsubscribeToPresence(sipUri);
			}

			m_presenceSubscriptionsBySipUri = new Dictionary<string, PresenceSubscription>();
		}
	}
}

