﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using log4net;
using Lync.Widget.Service.Common;
using Microsoft.Rtc.Collaboration;
using Microsoft.Rtc.Collaboration.Presence;
using Microsoft.Rtc.Signaling;

namespace Lync.Widget.Service.Presence
{
	/// <summary>
	/// The default UCMA implementation of IPresenceProvider.
	/// Creates a collaboration platfrom and endpoint and then subscribes to the presence updates of the requested contacts
	/// </summary>
	public class UcPresenceProvider : IPresenceProvider
	{
		private static readonly ILog Log = LogManager.GetLogger(MethodBase.GetCurrentMethod().DeclaringType);

		private ApplicationEndpoint m_applicationEndpoint;
		private RemotePresenceView m_remotePresenceView;

		public event EventHandler<PresenceChangedEventArgs> PresenceChanged;

		public UcPresenceProvider(ApplicationEndpoint endpoint)
		{
			m_applicationEndpoint = endpoint;
		}

		public bool Start()
		{
			return InitializeRemotePresenceView();
		}

		public void SubscribeToPresence(string sipUri)
		{
			if (!IsAlreadySubscribed(sipUri))
			{
				Subscribe(sipUri);
			}
		}

		public void UnsubscribeToPresence(string sipUri)
		{
			if (IsAlreadySubscribed(sipUri))
			{
				Unsubscribe(sipUri);
			}
		}

		private bool IsAlreadySubscribed(string sipUri)
		{
			if (m_remotePresenceView != null)
			{
				var subscriptions = m_remotePresenceView.GetPresentities();
				return subscriptions.Any(s => s.Equals(sipUri, StringComparison.InvariantCultureIgnoreCase));
			}
			else return false;
		}

		private void Subscribe(string sipUri)
		{
			SipUriParser uriParser;
			if (SipUriParser.TryParse(sipUri, out uriParser))
			{
				var subscriptionTargets = new List<RemotePresentitySubscriptionTarget>();
				subscriptionTargets.Add(new RemotePresentitySubscriptionTarget(uriParser.ToString()));
				// Immediately fires NotificationReceived with current presence of targets
				m_remotePresenceView.StartSubscribingToPresentities(subscriptionTargets);
			}
		}

		private void Unsubscribe(string sipUri)
		{
			SipUriParser uriParser;
			if (SipUriParser.TryParse(sipUri, out uriParser))
			{
				m_remotePresenceView.StartUnsubscribingToPresentities(new List<string> { sipUri });
			}
		}

		private void RemotePresenceView_PresenceNotificationReceived(object sender, RemotePresentitiesNotificationEventArgs e)
		{
			List<PresenceSubscription> presences = new List<PresenceSubscription>();
			foreach (var notification in e.Notifications)
			{
				presences.Add(ConvertNotificationToPresence(notification));
			}
			if (PresenceChanged != null)
			{
				PresenceChanged(this, new PresenceChangedEventArgs { Presences = presences });
			}
		}

		private PresenceSubscription ConvertNotificationToPresence(RemotePresentityNotification notification)
		{
			//after the first notification for a contact, you only get the categories that changed next time
			//so after the first notification if presence changes, you only get presence; if name changes you only get contact card
			long availability = PresenceSubscription.NoChangeInAvailability;
			string status = PresenceSubscription.NoChange;
			string contactName = PresenceSubscription.NoChange;
			string emailAddress = PresenceSubscription.NoChange;
			string jobTitle = PresenceSubscription.NoChange;
			string phone = PresenceSubscription.NoChange;
			string mobile = PresenceSubscription.NoChange;

			var presence = notification.AggregatedPresenceState;
			if (presence != null)
			{
				availability = presence.AvailabilityValue;
				status = AvailabilityToStatusConverter.Convert(presence.AvailabilityValue);
				if (presence.Activity != null)
				{
					switch (presence.Activity.ActivityToken)
					{
						case "in-a-meeting":
							status = "In a meeting";
							break;
						case "on-the-phone":
							status = "On the phone";
							break;
						case "in-a-conference":
							status = "In a conference call";
							break;

					}
				}
			}

			var contactCard = notification.ContactCard;
			if (contactCard != null)
			{
				contactName = contactCard.DisplayName;
				emailAddress = contactCard.EmailAddress;
				jobTitle = contactCard.JobTitle;
				var workphone = contactCard.PhoneNumbers.SingleOrDefault(p => p.PhoneType == "work");
				if (workphone != null) phone = workphone.Uri.Substring(workphone.Uri.IndexOf(':') + 1);
				var mobilestr = contactCard.PhoneNumbers.SingleOrDefault(p => p.PhoneType == "mobile");
				if (mobilestr != null) mobile = mobilestr.Uri.Substring(mobilestr.Uri.IndexOf(':') + 1);
			}

			return new PresenceSubscription
					{
						Availability = availability,
						ContactName = contactName,
						SipUri = notification.PresentityUri,
						Status = status,
						Email = emailAddress,
						JobTitle = jobTitle,
						Phone = phone,
						Mobile = mobile,
					};
		}

		private bool InitializeRemotePresenceView()
		{
			try
			{
				// RemotePresenceView provides a unified way to request presence for a set of presentities
				m_remotePresenceView = new RemotePresenceView(
					m_applicationEndpoint,
					new RemotePresenceViewSettings()
						{
							SubscriptionMode = RemotePresenceViewSubscriptionMode.Persistent,
						});


				// Wire up NotificationReceived before adding targets
				m_remotePresenceView.PresenceNotificationReceived +=
					new EventHandler<RemotePresentitiesNotificationEventArgs>(RemotePresenceView_PresenceNotificationReceived);
				Log.Info("UcPresenceProvider started");
				return true;
			}
			catch (Exception ex)
			{
				Log.Error("UCPresenceProvider failed: Could not initialize remote presence view {0}", ex);
				return false;
			}
		}
	}
}
